from discord.ext import commands
import discord

class Router():

    def __init__(self, bot):
        self.bot = bot

    def route(self, command: list, **kwargs):
        """Register command route."""
        def deco(func):
            return self.get_command_group(command, func, **kwargs)
        return deco

    def get_command_group(self, path: list, func = None, **kwargs):
        """Get command group."""
        if len(path) == 0:
            return self.bot
        # Try and find group
        group = self.bot.get_command(" ".join(path))
        # If it doesn't exist, create it group
        if group is None:
            # Get the parent group
            parent = self.get_command_group(path[:-1])
            # Create the func if it doesn't exist
            if func is None:
                async def func(ctx):
                    raise commands.CommandNotFound(f'Command "{path[-1]}" is not found')
            # Create the group
            group = commands.Group(func, name=path[-1], invoke_without_command=True, **kwargs)
            parent.add_command(group)
        # If it does exist, but needs to be overwritten, update callback
        elif func:
            group.callback = func

        return group



from enum import Enum
import discord


class Alert(discord.Embed):
    """ A subclass of discord.Embed for stylish alert messages.

        :param style: A contextual class describing the style of the alert.
            There are 8 main alert styles, but you can also provide your own style in the form of a (color, emoji) tuple.
            Primary and Secondary styles use custom emoji, so are unable to be used out of the box.
            To work around this, I've provided the source .svgs in the repository for the custom emojis used throughout the project. you can add these to a server that your bot is in, then create a custom style.
            Alternatively, if you want, you can support me on [ko-fi](https://ko-fi.com/s/7705c20532), and I'll invite your bot to my server where it'll have access to the original custom emojis.
        :param title: The title of the alert, will be wrapped in emoji and alert name unless specified in options.
        :param description: An optional description of the alert, use your imagination for it's use.
        :param **options: Alert options to customize it's look.
            - `emoji`: Defaults to `True`. Can be set to false to remove the emoji from the alert title.
            - `name`: Defaults to `True`. Can be set to false to remove the name of the alert from the title.
    """

    class Style(Enum):
        PRIMARY = (0x7289DA, "<:primary:808874731763007488>")
        SECONDARY = (0x99AAB5, "<:secondary:808874731758813205>")
        SUCCESS = (0x77B255, "\u2705")
        DANGER = (0xDD2E44, "\U0001f6ab")
        WARNING = (0xFFCC4D, "\u26a0\ufe0f")
        INFO = (0x3B88C3, "\u2139\ufe0f")
        LIGHT = (0xE6E7E8, "\U0001f533")
        DARK = (0x31373D, "\U0001f532")

        def __getitem__(self, key):
            return self.value[key]

    def __init__(self, style: Style, title: str, description: str = discord.Embed.Empty, **options):
        super().__init__(
            color=style[0],
            title=self.process_title(style, title, **options),
            description=description
        )

    @classmethod
    def process_title(cls, style: Style, title: str, **options):
        output: str = ''

        if options.get("emoji") is not False:
            output += style[1] + " "

        if (name := options.get("name", style.name.capitalize())) is not False:
            output += f"`{name}:` "

        return output + f"**{title}**"



import discord
from discord.ext import commands


class HelpCommand(commands.DefaultHelpCommand):
    """Set up help command for the bot."""

    def __init__(self, color, **options):
        super().__init__(**options)
        self.color = color

    async def send_bot_help(self, mapping):
        """Send bot command page."""
        commands = {k: mapping[k] for k in [None, *list(mapping.keys())[:-1]]}

        embed = self.create_embed(
            title=self.sig('help'),
            description=f"Welcome to the {self.context.bot.user.name} help command!",
            fields=[{
                    "name": self.sig(cog),
                    "value": await self.list_commands(cog_commands)
            } for cog, cog_commands in commands.items() if await self.filter_commands(cog_commands)]
        )
        await self.get_destination().send(embed=embed)

    async def send_cog_help(self, cog):
        """Send cog command page."""
        embed = self.create_embed(
            title=self.sig(cog),
            description=cog.description,
            **({"fields": [{
                "name": "Commands:",
                "value": await self.list_commands(cog.get_commands())
            }]} if cog.get_commands() else {})
        )
        await self.get_destination().send(embed=embed)

    async def send_group_help(self, group):
        """Send command group page."""
        embed = self.create_embed(
            title=self.sig(group, sig=True),
            description=group.help,
            **({"fields": [{
                "name": "Subcommands:",
                "value": await self.list_commands(group.commands)
            }]} if group.commands else {})
        )
        await self.get_destination().send(embed=embed)

    async def send_command_help(self, command):
        """Send command page."""
        embed = self.create_embed(
            title=self.sig(command, sig=True),
            description=command.help,
        )
        await self.get_destination().send(embed=embed)

    async def command_not_found(self, string):
        """Returns message when command is not found."""
        sig = self.sig(string)
        return f"Command {sig} is not found."

    async def subcommand_not_found(self, command, string):
        """Returns message when subcommand is not found."""
        sig = self.sig(command)
        if isinstance(command, commands.Group) and len(command.all_commands) > 0:
            return f"Command {sig} has no subcommand named `{string}`."
        else:
            return f"Command {sig} has no subcommands."

    async def send_error_message(self, error):
        """Send error message, override to support sending embeds."""
        await self.get_destination().send(
            embed=self.create_embed(title="Command Not Found", description=error))

    async def list_commands(self, commands):
        """Create a formatted string listing commands in a group."""
        return "\n".join([self.sig(command, doc=True) for command in await self.filter_commands(commands)])

    def create_embed(self, fields: list = (), **kwargs):
        """Create help embed."""
        embed = discord.Embed(color=self.color, **kwargs)
        for field in fields:
            embed.add_field(**field, inline=False)
        embed.set_footer(
            text=f"Type {self.clean_prefix}help command for more info on a command. You can also type {self.clean_prefix}help category for more info on a category.")
        return embed

    def sig(self, command, *, doc=False, sig=False):
        """List the command as a one-liner."""
        # Command is cog
        if isinstance(command, commands.Cog):
            return command.qualified_name
        if command is None:
            return '\u200B'
        # Command is regular
        display = self.get_command_signature(command).strip() if sig else f'{self.clean_prefix}{command}'
        return f'`{display}` {(command.short_doc if doc else "")}'



from __future__ import annotations
import re
import inspect
import typing as t
from abc import ABC
import discord
from discord.ext import commands
import asyncio

from blurple.ui import Alert


class Reply(ABC):
    """ Get a reply from the user.

        :param ctx: The discord.ext.commands.Context variable
        :param validate: An optional parameter to validate the reply.
            If left blank, no validation will be performed.
            If you pass a list/set, validation will succeed when the reply content is found inside the list/set.
            If you pass a string, validation will succeed when the reply content matches the string as a regex.
            If you pass a function or coroutine, the function will be called, and the coroutine awaited.
                Validation will succeed when the function returns a Truthy value.
                The reply object will be passed as a parameter.
        :param error: An optional parameter specifying the message to send when the user fails validation, defaults to a simple "Invalid Reply" alert.
        :return: A :Reply: object.
    """

    def __init__(self,
            ctx: commands.Context,
            *,
            validate: t.Optional[t.Union[str, t.Callable, t.List]] = None,
            on_error: t.Union[str, discord.Embed] = Alert(Alert.Style.DANGER, "Invalid Reply"),
            timeout = 180,
            **kwargs) -> None:
        """Set up the reply class instance

        .isTimeout
        .message
        .author
        .how_long_it_took_to_reply
        """
        # Set up
        self.ctx = ctx
        self.validate = validate
        self.on_error = on_error
        self.error = None
        self.timeout = timeout
        self.kwargs = kwargs

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return f"<{self.__class__.__name__} for '{self.event}'>"

    async def result(self):
        """Await the result of the reply."""
        await self.on_reply_init(**self.kwargs)  # Event method
        reply = await self.get_valid_reply()
        await self.cleanup()  # Event method
        return reply

    async def cleanup(self):
        """Clean up reply after result."""
        await self.on_reply_complete()
        await self.delete_error()

    async def get_valid_reply(self):
        """Wrap get_reply with validation, error handling, and recursive calls."""
        reply = await self.get_reply()

        if reply is not None:  # Reply hasn't timed out
            # Validate reply
            is_valid = await self.validate_reply(reply, self.validate)
            # If reply isn't valid, recursively call function
            if not is_valid:
                await self.send_error()
                return await self.get_valid_reply()

        return reply

    async def get_reply(self):
        """Get a reply from the user, no validation."""
        await self.on_pre_reply()  # Event method
        # Wait for reply
        try:
            raw_reply = await self.ctx.bot.wait_for(
                self.event,
                check=self.reply_check,
                timeout=self.timeout
            )
        except asyncio.TimeoutError:
            reply = None
        else:
            r = await self.on_reply_attempt(raw_reply)  # Event method
            reply = r if r else raw_reply
        return reply

    async def send_error(self) -> discord.Message:
        """ Send an error message to the user.

            Will replace the current error message.
            :param error: An embed or a string representing the error message.
        """
        await self.delete_error()
        if isinstance(self.on_error, discord.Embed):
            self.error = await self.ctx.send(embed=self.on_error)
        elif isinstance(self.on_error, str):
            self.error = await self.ctx.send(self.on_error)
        return self.error

    async def delete_error(self) -> None:
        """Delete the current error message, if it exists."""
        if self.error is None:
            return
        await self.error.delete()
        self.error = None


    @classmethod
    async def validate_reply(cls, reply, valid) -> bool:
        if valid is None:
            return True
        content = cls.get_reply_content(reply)
        if isinstance(valid, str):
            return bool(re.search(valid, content))
        if cls.is_container(valid):
            return content in valid
        if callable(valid):
            if inspect.iscoroutinefunction(object):
                return await valid()
            return valid()

    @staticmethod
    def get_reply_content(reply):
        """Retrieve the content of the reply."""
        if isinstance(reply, discord.Message):
            return reply.content
        if isinstance(reply, (discord.Reaction, discord.RawReactionActionEvent)):
            return str(reply.emoji)

    @staticmethod
    def is_container(obj: t.Union[t.Container, t.Any]):
        return getattr(obj, "__contains__", False)

    @classmethod
    async def result_between(cls, replies: t.Container[Reply]):
        """ Return the first completed result between multiple reply objects.

            :param replies: A collection of Reply objects.
        """
        # Prepare tasks
        timeouts = []
        def parse_task(reply: Reply):
            # Handle timeout
            timeouts.append(reply.timeout)
            reply.timeout = None
            # Return task
            return asyncio.create_task(reply.result(), name=reply)

        # Wait tasks
        tasks = [parse_task(task) for task in replies]
        task, result = await cls.wait_tasks(tasks, timeout=min(timeouts))

        # Get original reply object
        for reply in replies:
            if str(reply) == task.get_name():
                break
        # Run cleanup on cancelled replies
        replies.remove(reply)
        for cancelled in replies:
            await cancelled.cleanup()

        # Return original reply object and the result
        return reply, result

    @staticmethod
    async def wait_tasks(tasks: t.Container[asyncio.Task], timeout: int) -> t.Tuple[t.Optional[asyncio.Future], t.Optional[t.Any]]:
        """ Try block to asyncio.wait a set of tasks with timeout handling.

            :param tasks: A collection of task objects
            :param timeout: How long in seconds to wait until a timeout occurs.
            :return: A tuple containing the task and the result. Both will be None if a timeout occurs.
        """
        done, pending = await asyncio.wait(tasks, timeout=timeout, return_when=asyncio.FIRST_COMPLETED)
        for rest in pending:
            rest.cancel()

        if done:
            task: asyncio.Future = done.pop()
            return task, task.result()

        return None, None

    async def on_reply_init(self):
        """Runs on reply class creation."""

    async def on_pre_reply(self):
        """Runs before wait_for event."""

    def reply_check(self, reply):
        """The check to authorize reply."""

    async def on_reply_attempt(self, reply):
        """ Any actions to do after each reply attempt, can be run multiple times if validation is on.

            :return: You can optionally return a parsed version of the reply to be used instead of the raw reply object.
        """

    async def on_reply_complete(self):
        """Any actions to do after reply is completely done, only runs once reply is validated and returned."""


class MessageReply(Reply):
    event = "message"

    def reply_check(self, reply: discord.Message):
        return reply.author.id == self.ctx.author.id and \
               reply.channel.id == self.ctx.channel.id and \
               not reply.content.startswith("\\")

    async def on_reply_attempt(self, reply: discord.Message):
        await reply.delete()


class ReactionAddReply(Reply):
    event = "raw_reaction_add"

    async def on_reply_init(self, message):
        self.message = message
        if self.is_container(self.validate):
            for react in self.validate:
                await self.message.add_reaction(react)

    def reply_check(self, payload):
        return payload.user_id == self.ctx.author.id and \
               payload.message_id == self.message.id

    async def on_reply_attempt(self, payload: discord.RawReactionActionEvent):
        await self.message.remove_reaction(payload.emoji, self.ctx.bot.get_user(payload.user_id))
        return payload

    async def on_reply_complete(self):
        await self.message.clear_reactions()


class ReactionRemoveReply(ReactionAddReply):
    event = "raw_reaction_remove"

    async def on_reply_attempt(self, payload: discord.RawReactionActionEvent):
        return payload
